#!/usr/bin/env python
"""
Median absolute deviation test

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2011-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   mad

History
    * Written Nov 2011 by Matthias Cuntz - mc (at) macu (dot) de
    * ND-array, act on axis=0, May 2012, Matthias Cuntz
    * Removed bug in broadcasting, axis=0 did not always work: spread md and
      MAD to input dimensions, Jun 2012, Matthias Cuntz
    * Better usage of numpy possibilities, e.g. using np.diff,
      remove spreads, Jun 2012, Matthias Cuntz
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Use bottleneck for medians, otherwise loop over axis=1,
      Jul 2013, Matthias Cuntz and Juliane Mai
    * Re-allow masked arrays and arrays with NaNs, Jul 2013, Matthias Cuntz
    * Removed bug in NaN treatment, Oct 2013, Matthias Cuntz
    * Keyword nonzero, Oct 2013, Matthias Cuntz
    * Using numpy docstring format, May 2020, Matthias Cuntz
    * Code refactoring, Sep 2021, Matthias Cuntz
    * Ported to pyjams, Jan 2022, Matthias Cuntz
    * Only one test of bottleneck availability, Jan 2022, Matthias Cuntz
    * Return all False instead of all True if all masked before MAD started,
      Jan 2022, Matthias Cuntz

"""
import numpy as np


__all__ = ['mad']


def mad(datin, z=7, deriv=0, nozero=False):
    """
    Median absolute deviation test

    The test acts either on raw values, or on 1st or 2nd derivatives.

    Return mask with True where value is out of range, i.e.
    :math:`< (median - z.MAD/0.6745)` or :math:`> (md + z.MAD/0.6745)`.

    Parameters
    ----------
    datin : array or masked array
        `mad` acts on ``axis=0``
    z : float, optional
        Input is allowed to deviate maximum *z* (estimators of) standard
        deviations from the median (default: 7)
    deriv : int, optional
        Act on raw input (0, default), on first derivatives (1), or on 2nd
        derivatives (2)
    nozero : bool, optional
        Exclude zeros (0.) from input *datin* if True.

    Returns
    -------
    array of bool
        False everywhere except where input deviates more than
        *z* standard deviations from median

    Notes
    -----
    If input is an nd-array then mad is checked along the first axis for
    outliers.

    1st derivative is calculated simply as ``d = datin[1:n]-datin[0:n-1]``
    because mean of left and right difference would give 0 for spikes.

    If ``all(d.mask==True)`` then return ``d.mask``, which is all True.

    NaN does not return True because this would remove points adjacent to NaN
    if ``deriv>0``.

    Examples
    --------
    >>> import numpy as np
    >>> y = np.array([-0.25, 0.68, 0.94, 1.15, 2.26, 2.35, 2.37, 2.40, 2.47,
    ...               2.54, 2.62, 2.64, 2.90, 2.92, 2.92, 2.93, 3.21, 3.26,
    ...               3.30, 3.59, 3.68, 4.30, 4.64, 5.34, 5.42, 8.01])

    MAD on raw data

    >>> print(mad(y))
    [False False False False False False False False False False False False
     False False False False False False False False False False False False
     False False]
    >>> print(mad(y, z=4))
    [False False False False False False False False False False False False
     False False False False False False False False False False False False
     False  True]
    >>> print(mad(y, z=3))
    [ True False False False False False False False False False False False
     False False False False False False False False False False False False
      True  True]

    MAD on 2nd derivatives

    >>> print(mad(y, z=4, deriv=2))
    [False False False False False False False False False False False False
     False False False False False False False False False False False  True]

    Use for masking arrays

    >>> my = np.ma.array(y, mask=mad(y, z=4))
    >>> print(my)
    [-0.25 0.68 0.94 1.15 2.26 2.35 2.37 2.4 2.47 2.54 2.62 2.64 2.9 2.92 2.92
     2.93 3.21 3.26 3.3 3.59 3.68 4.3 4.64 5.34 5.42 --]

    MAD on several dimensions

    >>> yy = np.transpose(np.array([y, y]))
    >>> print(np.transpose(mad(yy, z=4)))
    [[False False False False False False False False False False False False
      False False False False False False False False False False False False
      False  True]
     [False False False False False False False False False False False False
      False False False False False False False False False False False False
      False  True]]
    >>> yyy = np.transpose(np.array([y, y, y]))
    >>> print(np.transpose(mad(yyy, z=3)))
    [[ True False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]
     [ True False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]
     [ True False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]]

    Masked arrays

    >>> my = np.ma.array(y, mask=np.zeros(y.shape))
    >>> my.mask[-1] = True
    >>> print(mad(my, z=4))
    [True False False False False False False False False False False False
     False False False False False False False False False False False False
     False --]
    >>> print(mad(my, z=3))
    [True False False False False False False False False False False False
     False False False False False False False False False False False True
     True --]

    Arrays with NaNs

    >>> ny = y.copy()
    >>> ny[-1] = np.nan
    >>> print(mad(ny, z=4))
    [ True False False False False False False False False False False False
     False False False False False False False False False False False False
     False False]
    >>> print(mad(ny, z=3))
    [ True False False False False False False False False False False False
     False False False False False False False False False False False  True
      True False]

    Exclude zeros

    >>> zy = y.copy()
    >>> zy[1] = 0.
    >>> print(mad(zy, z=3))
    [ True  True False False False False False False False False False False
     False False False False False False False False False False False False
      True  True]
    >>> print(mad(zy, z=3, nozero=True))
    [ True False False False False False False False False False False False
     False False False False False False False False False False False False
      True  True]

    """
    if nozero:
        idatin = datin.copy()
        ii = np.where(idatin == 0.)[0]
        if ii.size > 0:
            idatin[ii] = np.nan
    else:
        idatin = datin
    sn = list(np.shape(idatin))
    n  = sn[0]
    if deriv == 0:
        m      = n
        d      = idatin
    elif deriv == 1:
        m      = n-1
        sm     = sn
        sm[0]  = m
        d      = np.diff(idatin, axis=0)
    elif deriv == 2:
        m      = n-2
        sm     = sn
        sm[0]  = m
        d      = np.diff(idatin, n=2, axis=0)
    else:
        raise ValueError('Unimplemented option.')

    # Shortcut if all masked
    ismasked = isinstance(d, np.ma.MaskedArray)
    if not ismasked:
        ii = np.where(~np.isfinite(d))[0]
        d  = np.ma.array(d)
        if ii.size > 0:
            d[ii] = np.ma.masked

    if np.all(d.mask):
        if ismasked:
            return d.mask
        else:
            return np.zeros(d.shape, dtype=bool)

    # Use bottleneck if available
    try:     # pragma: no cover
        import bottleneck as bn
        med = bn.median
    except:  # pragma: no cover
        med = np.median

    # Median
    oldsettings = np.geterr()
    np.seterr(invalid='ignore')
    if d.ndim == 1:
        dd = d.compressed()
        md = med(dd)
        # Median absolute deviation
        MAD = med(np.abs(dd-md))
        # Range around median
        thresh = MAD * (z/0.6745)
        # True where outside z-range
        res = (d < (md-thresh)) | (d > (md+thresh))
    elif d.ndim == 2:
        res = np.empty(d.shape, dtype=bool)
        for i in range(d.shape[1]):
            di = d[:, i]
            dd = di.compressed()
            md = med(dd)
            # Median absolute deviation
            MAD = med(np.abs(dd-md))
            # Range around median
            thresh = MAD * (z/0.6745)
            # True where outside z-range
            res[:, i] = (d[:, i] < (md-thresh)) | (d[:, i] > (md+thresh))
    else:
        np.seterr(**oldsettings)
        raise ValueError('datin.ndim must be <= 2')

    np.seterr(**oldsettings)
    if ismasked:
        return res
    else:
        # got masked because of NaNs
        if isinstance(res, np.ma.MaskedArray):
            return np.where(res.mask, False, res)
        else:
            return res


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import numpy as np
    # y = np.array([-0.25,0.68,0.94,1.15,2.26,2.35,2.37,2.40,2.47,2.54,2.62,
    #                2.64,2.90,2.92,2.92,2.93,3.21,3.26,3.30,3.59,3.68,4.30,
    #                4.64,5.34,5.42,8.01],dtype=float)
    # print(mad(y))
    # #[False False False False False False False False False False False False
    # # False False False False False False False False False False False False
    # # False False]

    # print(mad(y,z=4))
    # #[False False False False False False False False False False False False
    # # False False False False False False False False False False False False
    # # False  True]

    # print(mad(y,z=3))
    # #[ True False False False False False False False False False False False
    # # False False False False False False False False False False False False
    # #  True  True]

    # print(mad(y,z=4,deriv=2))
    # #[False False False False False False False False False False False False
    # # False False False False False False False False False False False  True]

    # my = np.ma.array(y, mask=mad(y,z=4))
    # print(my)
    # #[-0.25 0.68 0.94 1.15 2.26 2.35 2.37 2.4 2.47 2.54 2.62 2.64 2.9 2.92 2.92
    # # 2.93 3.21 3.26 3.3 3.59 3.68 4.3 4.64 5.34 5.42 --]

    # yy = np.transpose(np.array([y,y]))
    # print(np.transpose(mad(yy,z=4)))
    # #[[False False False False False False False False False False False False
    # #  False False False False False False False False False False False False
    # #  False  True]
    # # [False False False False False False False False False False False False
    # #  False False False False False False False False False False False False
    # #  False  True]]

    # yyy = np.transpose(np.array([y,y,y]))
    # print(np.transpose(mad(yyy,z=3)))
    # #[[ True False False False False False False False False False False False
    # #  False False False False False False False False False False False False
    # #   True  True]
    # # [ True False False False False False False False False False False False
    # #  False False False False False False False False False False False False
    # #   True  True]
    # # [ True False False False False False False False False False False False
    # #  False False False False False False False False False False False False
    # #   True  True]]

    # my = np.ma.array(y, mask=np.zeros(y.shape))
    # my.mask[-1] = True
    # print(mad(my,z=4))
    # #[False False False False False False False False False False False False
    # # False False False False False False False False False False False False
    # # False --]

    # print(mad(my,z=3))
    # #[True False False False False False False False False False False False
    # # False False False False False False False False False False False True
    # # True --]

    # ny = y
    # ny[-1] = np.nan
    # print(mad(ny,z=4))
    # #[False False False False False False False False False False False False
    # # False False False False False False False False False False False False
    # # False False]

    # print(mad(ny,z=3))
    # #[True False False False False False False False False False False False
    # # False False False False False False False False False False False True
    # # True False]



#!/usr/bin/env python
"""
This is the unittest for esat module.

python -m unittest -v tests/test_air_humidity.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_air_humidity.py

"""
import unittest


class TestEsat(unittest.TestCase):
    """
    Tests for test_air_humidity.py
    """

    def test_esat(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import esat
        import warnings

        T0 = 273.15

        # scalar
        T1 = T0 + 20.
        assert isinstance(esat(T1), float)
        assert np.around(esat(T1), 3) == 2335.847
        T1 = T0 - 20.
        assert isinstance(esat(T1), float)
        assert np.around(esat(T1), 3) == 103.074

        # list - different formulas with different word cases
        T = [20., -20.]
        T1 = [ tt + T0 for tt in T ]
        assert isinstance(esat(T1), list)
        self.assertEqual(
            list(np.around(esat(T1), 3)),
            [2335.847, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='GoffGratch'), 3)),
            [2335.847, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='MartiMauersberger'), 3)),
            [2335.847, 103.650])
        self.assertEqual(
            list(np.around(esat(T1, formula='MagnusTeten'), 3)),
            [2335.201, 102.771])
        self.assertEqual(
            list(np.around(esat(T1, formula='buck'), 3)),
            [2338.340, 103.286])
        self.assertEqual(
            list(np.around(esat(T1, formula='Buck_original'), 3)),
            [2337.282, 103.267])
        self.assertEqual(
            list(np.around(esat(T1, formula='wmo'), 3)),
            [2337.080, 103.153])
        self.assertEqual(
            list(np.around(esat(T1, formula='WEXLER'), 3)),
            [2323.254, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='Sonntag'), 3)),
            [2339.249, 103.249])
        self.assertEqual(
            list(np.around(esat(T1, formula='Bolton'), 3)),
            [2336.947, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='Fukuta'), 3)),
            [2335.847, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='HylandWexler'), 3)),
            [2338.804, 103.260])
        self.assertEqual(
            list(np.around(esat(T1, formula='IAPWS'), 3)),
            [2339.194, 103.074])
        self.assertEqual(
            list(np.around(esat(T1, formula='MurphyKoop'), 3)),
            [2339.399, 103.252])

        # tuple
        T1 = tuple(T1)
        assert isinstance(esat(T1), tuple)
        # ndarray
        T  = np.array(T)
        T1 = T + T0
        assert isinstance(esat(T1), np.ndarray)
        # masked_array
        T1 = np.ma.array(T + T0, mask=(T == 20.))
        es = esat(T1)
        assert isinstance(es, np.ndarray)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(esat(T1), 3)),
            [masked, 103.074])

        # liquid
        T1 = [ tt + T0 for tt in T ]
        self.assertEqual(
            list(np.around(esat(T1, liquid=True), 3)),
            [2335.847, 125.292])
        self.assertEqual(
            list(np.around(esat(T1, formula='Fukuta', liquid=True), 3)),
            [2335.847, 125.079])

        # undef, scalar
        T1 = -9998.
        es = esat(T1, undef=-9998.)
        assert isinstance(es, float)
        self.assertEqual(es, -9998.)
        T1 = -9999.
        es = esat(T1)
        assert isinstance(es, float)
        self.assertEqual(es, -9999.)

        # undef, list
        T1 = [20. + T0, -9998.]
        es = esat(T1, undef=-9998.)
        assert isinstance(es, list)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9998.])
        T1 = [20. + T0, -9999.]
        es = esat(T1)
        assert isinstance(es, list)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9999.])

        # undef, tuple
        T1 = (20. + T0, -9998.)
        es = esat(T1, undef=-9998.)
        assert isinstance(es, tuple)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9998.])
        T1 = (20. + T0, -9999.)
        es = esat(T1)
        assert isinstance(es, tuple)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9999.])

        # undef, ndarray
        T1 = np.array([20. + T0, -9998.])
        es = esat(T1, undef=-9998.)
        assert isinstance(es, np.ndarray)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9998.])
        T1 = np.array([20. + T0, -9999.])
        es = esat(T1)
        assert isinstance(es, np.ndarray)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, -9999.])

        # undef, masked array
        T = np.array([20. + T0, -9998., -9999.])
        T1 = np.ma.array(T, mask=(T == -9999.))
        es = esat(T1, undef=-9998.)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, masked, masked])
        T = np.array([20. + T0, -9998., -9999.])
        T1 = np.ma.array(T, mask=(T == -9998.))
        es = esat(T1)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(es, 3)),
            [2335.847, masked, masked])

        # errors
        # T < 0
        T1 = T0 - T0
        self.assertRaises(AssertionError, esat, T1)
        # T > 373.15
        T1 = T0 + 101.
        self.assertRaises(AssertionError, esat, T1)
        # wrong formula
        T1 = T0 + 20.
        self.assertRaises(ValueError, esat, T1, formula='dummy')
        T1 = T0 - 20.
        self.assertRaises(ValueError, esat, T1, formula='dummy')

        # warning T < 100
        T1 = T0 - 200.
        with warnings.catch_warnings(record=True) as w:
            es = esat(T1)
            assert len(w) > 0

    def test_eair2rhair(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import eair2rhair

        eair = 1000.
        Ta = 293.15
        res2 = 42.81

        # scalar
        e1 = eair
        T1 = Ta
        assert isinstance(eair2rhair(e1, T1), float)
        assert np.around(eair2rhair(e1, T1) * 100., 2) == res2

        # list
        e1 = [eair, eair]
        T1 = Ta
        assert isinstance(eair2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, res2])
        e1 = eair
        T1 = [Ta, Ta]
        assert isinstance(eair2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, res2])
        e1 = [eair, eair]
        T1 = [Ta, Ta]
        assert isinstance(eair2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        T1 = tuple(T1)
        assert isinstance(eair2rhair(e1, T1), tuple)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        T1 = np.array(T1)
        assert isinstance(eair2rhair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(eair2rhair(e1, T1) * 100., 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        T1 = np.ma.array(T1)
        assert isinstance(eair2rhair(e1, T1), np.ndarray)
        assert isinstance(eair2rhair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2rhair(e1, T1) * 100., 2)),
            [res2, res2])

        # mixed types
        e1 = [eair, eair]
        T1 = tuple([Ta, Ta])
        assert isinstance(eair2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, res2])
        e1 = tuple([eair, eair])
        T1 = [Ta, Ta]
        assert isinstance(eair2rhair(e1, T1), tuple)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, res2])
        e1 = np.array(e1)
        T1 = [Ta, Ta]
        assert isinstance(eair2rhair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(eair2rhair(e1, T1) * 100., 2)),
            [res2, res2])
        e1 = [eair, eair]
        T1 = np.array(T1)
        assert isinstance(eair2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        T1 = Ta
        assert isinstance(eair2rhair(e1, T1, undef=-9998.), float)
        assert np.around(eair2rhair(e1, T1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        T1 = Ta
        assert isinstance(eair2rhair(e1, T1), float)
        assert np.around(eair2rhair(e1, T1), 2) == -9999.
        e1 = eair
        T1 = -9999.
        assert isinstance(eair2rhair(e1, T1), float)
        assert np.around(eair2rhair(e1, T1), 2) == -9999.

        # undef, list
        e1 = [eair, -9999.]
        T1 = [Ta, Ta]
        assert isinstance(eair2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, -9999. * 100.])
        e1 = [eair, eair]
        T1 = [Ta, -9999.]
        assert isinstance(eair2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in eair2rhair(e1, T1)], 2)),
            [res2, -9999. * 100.])

        # undef, masked array
        e1 = np.ma.array([eair, -9998., -9999.])
        T1 = [Ta, Ta, Ta]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(eair2rhair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2rhair(e1, T1) * 100., 2)),
            [res2, masked, masked])
        e1 = [eair, eair, eair]
        T1 = np.ma.array([Ta, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(eair2rhair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2rhair(e1, T1) * 100., 2)),
            [res2, masked, masked])

    def test_rhair2eair(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import rhair2eair

        rhair = 0.5
        Ta = 293.15
        res2 = 1167.92

        # scalar
        e1 = rhair
        T1 = Ta
        assert isinstance(rhair2eair(e1, T1), float)
        assert np.around(rhair2eair(e1, T1), 2) == res2

        # list
        e1 = [rhair, rhair]
        T1 = Ta
        assert isinstance(rhair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = rhair
        T1 = [Ta, Ta]
        assert isinstance(rhair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = [rhair, rhair]
        T1 = [Ta, Ta]
        assert isinstance(rhair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        T1 = tuple(T1)
        assert isinstance(rhair2eair(e1, T1), tuple)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        T1 = np.array(T1)
        assert isinstance(rhair2eair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        T1 = np.ma.array(T1)
        assert isinstance(rhair2eair(e1, T1), np.ndarray)
        assert isinstance(rhair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])

        # mixed types
        e1 = [rhair, rhair]
        T1 = tuple([Ta, Ta])
        assert isinstance(rhair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = tuple([rhair, rhair])
        T1 = [Ta, Ta]
        assert isinstance(rhair2eair(e1, T1), tuple)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = np.array(e1)
        T1 = [Ta, Ta]
        assert isinstance(rhair2eair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = [rhair, rhair]
        T1 = np.array(T1)
        assert isinstance(rhair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        T1 = Ta
        assert isinstance(rhair2eair(e1, T1, undef=-9998.), float)
        assert np.around(rhair2eair(e1, T1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        T1 = Ta
        assert isinstance(rhair2eair(e1, T1), float)
        assert np.around(rhair2eair(e1, T1), 2) == -9999.
        e1 = rhair
        T1 = -9999.
        assert isinstance(rhair2eair(e1, T1), float)
        assert np.around(rhair2eair(e1, T1), 2) == -9999.

        # undef, list
        e1 = [rhair, -9999.]
        T1 = [Ta, Ta]
        assert isinstance(rhair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, -9999.])
        e1 = [rhair, rhair]
        T1 = [Ta, -9999.]
        assert isinstance(rhair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, -9999.])

        # undef, masked array
        e1 = np.ma.array([rhair, -9998., -9999.])
        T1 = [Ta, Ta, Ta]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(rhair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, masked, masked])
        e1 = [rhair, rhair, rhair]
        T1 = np.ma.array([Ta, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(rhair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(rhair2eair(e1, T1), 2)),
            [res2, masked, masked])

    def test_eair2vpd(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import eair2vpd

        eair = 1000.
        Ta = 293.15
        res2 = 1335.85

        # scalar
        e1 = eair
        T1 = Ta
        assert isinstance(eair2vpd(e1, T1), float)
        assert np.around(eair2vpd(e1, T1), 2) == res2

        # list
        e1 = [eair, eair]
        T1 = Ta
        assert isinstance(eair2vpd(e1, T1), list)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])
        e1 = eair
        T1 = [Ta, Ta]
        assert isinstance(eair2vpd(e1, T1), list)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])
        e1 = [eair, eair]
        T1 = [Ta, Ta]
        assert isinstance(eair2vpd(e1, T1), list)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        T1 = tuple(T1)
        assert isinstance(eair2vpd(e1, T1), tuple)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        T1 = np.array(T1)
        assert isinstance(eair2vpd(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        T1 = np.ma.array(T1)
        assert isinstance(eair2vpd(e1, T1), np.ndarray)
        assert isinstance(eair2vpd(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])

        # mixed types
        e1 = [eair, eair]
        T1 = tuple([Ta, Ta])
        assert isinstance(eair2vpd(e1, T1), list)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])
        e1 = tuple([eair, eair])
        T1 = [Ta, Ta]
        assert isinstance(eair2vpd(e1, T1), tuple)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])
        e1 = np.array(e1)
        T1 = [Ta, Ta]
        assert isinstance(eair2vpd(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])
        e1 = [eair, eair]
        T1 = np.array(T1)
        assert isinstance(eair2vpd(e1, T1), list)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        T1 = Ta
        assert isinstance(eair2vpd(e1, T1, undef=-9998.), float)
        assert np.around(eair2vpd(e1, T1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        T1 = Ta
        assert isinstance(eair2vpd(e1, T1), float)
        assert np.around(eair2vpd(e1, T1), 2) == -9999.
        e1 = eair
        T1 = -9999.
        assert isinstance(eair2vpd(e1, T1), float)
        assert np.around(eair2vpd(e1, T1), 2) == -9999.

        # undef, list
        e1 = [eair, -9999.]
        T1 = [Ta, Ta]
        assert isinstance(eair2vpd(e1, T1), list)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, -9999.])
        e1 = [eair, eair]
        T1 = [Ta, -9999.]
        assert isinstance(eair2vpd(e1, T1), list)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, -9999.])

        # undef, masked array
        e1 = np.ma.array([eair, -9998., -9999.])
        T1 = [Ta, Ta, Ta]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(eair2vpd(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, masked, masked])
        e1 = [eair, eair, eair]
        T1 = np.ma.array([Ta, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(eair2vpd(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2vpd(e1, T1), 2)),
            [res2, masked, masked])

    def test_vpd2eair(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import vpd2eair

        vpd = 1000.
        Ta = 293.15
        res2 = 1335.85

        # scalar
        e1 = vpd
        T1 = Ta
        assert isinstance(vpd2eair(e1, T1), float)
        assert np.around(vpd2eair(e1, T1), 2) == res2

        # list
        e1 = [vpd, vpd]
        T1 = Ta
        assert isinstance(vpd2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])
        e1 = vpd
        T1 = [Ta, Ta]
        assert isinstance(vpd2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])
        e1 = [vpd, vpd]
        T1 = [Ta, Ta]
        assert isinstance(vpd2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        T1 = tuple(T1)
        assert isinstance(vpd2eair(e1, T1), tuple)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        T1 = np.array(T1)
        assert isinstance(vpd2eair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        T1 = np.ma.array(T1)
        assert isinstance(vpd2eair(e1, T1), np.ndarray)
        assert isinstance(vpd2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])

        # mixed types
        e1 = [vpd, vpd]
        T1 = tuple([Ta, Ta])
        assert isinstance(vpd2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])
        e1 = tuple([vpd, vpd])
        T1 = [Ta, Ta]
        assert isinstance(vpd2eair(e1, T1), tuple)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])
        e1 = np.array(e1)
        T1 = [Ta, Ta]
        assert isinstance(vpd2eair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])
        e1 = [vpd, vpd]
        T1 = np.array(T1)
        assert isinstance(vpd2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        T1 = Ta
        assert isinstance(vpd2eair(e1, T1, undef=-9998.), float)
        assert np.around(vpd2eair(e1, T1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        T1 = Ta
        assert isinstance(vpd2eair(e1, T1), float)
        assert np.around(vpd2eair(e1, T1), 2) == -9999.
        e1 = vpd
        T1 = -9999.
        assert isinstance(vpd2eair(e1, T1), float)
        assert np.around(vpd2eair(e1, T1), 2) == -9999.

        # undef, list
        e1 = [vpd, -9999.]
        T1 = [Ta, Ta]
        assert isinstance(vpd2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, -9999.])
        e1 = [vpd, vpd]
        T1 = [Ta, -9999.]
        assert isinstance(vpd2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, -9999.])

        # undef, masked array
        e1 = np.ma.array([vpd, -9998., -9999.])
        T1 = [Ta, Ta, Ta]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(vpd2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, masked, masked])
        e1 = [vpd, vpd, vpd]
        T1 = np.ma.array([Ta, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(vpd2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(vpd2eair(e1, T1), 2)),
            [res2, masked, masked])

    def test_rhair2vpd(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import rhair2vpd

        rhair = 0.5
        Ta = 293.15
        res2 = 1167.92

        # scalar
        rhair1 = rhair
        T1 = Ta
        assert isinstance(rhair2vpd(rhair1, T1), float)
        assert np.around(rhair2vpd(rhair1, T1), 2) == res2

        # list
        rhair1 = [rhair, rhair]
        T1 = Ta
        assert isinstance(rhair2vpd(rhair1, T1), list)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])
        rhair1 = rhair
        T1 = [Ta, Ta]
        assert isinstance(rhair2vpd(rhair1, T1), list)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])
        rhair1 = [rhair, rhair]
        T1 = [Ta, Ta]
        assert isinstance(rhair2vpd(rhair1, T1), list)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])

        # tuple
        rhair1 = tuple(rhair1)
        T1 = tuple(T1)
        assert isinstance(rhair2vpd(rhair1, T1), tuple)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])

        # ndarray
        rhair1 = np.array(rhair1)
        T1 = np.array(T1)
        assert isinstance(rhair2vpd(rhair1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])

        # masked_array
        rhair1 = np.ma.array(rhair1)
        T1 = np.ma.array(T1)
        assert isinstance(rhair2vpd(rhair1, T1), np.ndarray)
        assert isinstance(rhair2vpd(rhair1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])

        # mixed types
        rhair1 = [rhair, rhair]
        T1 = tuple([Ta, Ta])
        assert isinstance(rhair2vpd(rhair1, T1), list)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])
        rhair1 = tuple([rhair, rhair])
        T1 = [Ta, Ta]
        assert isinstance(rhair2vpd(rhair1, T1), tuple)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])
        rhair1 = np.array(rhair1)
        T1 = [Ta, Ta]
        assert isinstance(rhair2vpd(rhair1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])
        rhair1 = [rhair, rhair]
        T1 = np.array(T1)
        assert isinstance(rhair2vpd(rhair1, T1), list)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, res2])

        # undef, scalar
        rhair1 = -9998.
        T1 = Ta
        assert isinstance(rhair2vpd(rhair1, T1, undef=-9998.), float)
        assert np.around(rhair2vpd(rhair1, T1, undef=-9998.), 2) == -9998.
        rhair1 = -9999.
        T1 = Ta
        assert isinstance(rhair2vpd(rhair1, T1), float)
        assert np.around(rhair2vpd(rhair1, T1), 2) == -9999.
        rhair1 = rhair
        T1 = -9999.
        assert isinstance(rhair2vpd(rhair1, T1), float)
        assert np.around(rhair2vpd(rhair1, T1), 2) == -9999.

        # undef, list
        rhair1 = [rhair, -9999.]
        T1 = [Ta, Ta]
        assert isinstance(rhair2vpd(rhair1, T1), list)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, -9999.])
        rhair1 = [rhair, rhair]
        T1 = [Ta, -9999.]
        assert isinstance(rhair2vpd(rhair1, T1), list)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, -9999.])

        # undef, masked array
        rhair1 = np.ma.array([rhair, -9998., -9999.])
        T1 = [Ta, Ta, Ta]
        rhair1 = np.ma.array(rhair1, mask=(rhair1 == -9998.))
        assert isinstance(rhair2vpd(rhair1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, masked, masked])
        rhair1 = [rhair, rhair, rhair]
        T1 = np.ma.array([Ta, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(rhair2vpd(rhair1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(rhair2vpd(rhair1, T1), 2)),
            [res2, masked, masked])

    def test_vpd2rhair(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import vpd2rhair

        vpd = 1000.
        Ta = 293.15
        res2 = 57.19

        # scalar
        e1 = vpd
        T1 = Ta
        assert isinstance(vpd2rhair(e1, T1), float)
        assert np.around(vpd2rhair(e1, T1) * 100., 2) == res2

        # list
        e1 = [vpd, vpd]
        T1 = Ta
        assert isinstance(vpd2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, res2])
        e1 = vpd
        T1 = [Ta, Ta]
        assert isinstance(vpd2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, res2])
        e1 = [vpd, vpd]
        T1 = [Ta, Ta]
        assert isinstance(vpd2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        T1 = tuple(T1)
        assert isinstance(vpd2rhair(e1, T1), tuple)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        T1 = np.array(T1)
        assert isinstance(vpd2rhair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(vpd2rhair(e1, T1) * 100., 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        T1 = np.ma.array(T1)
        assert isinstance(vpd2rhair(e1, T1), np.ndarray)
        assert isinstance(vpd2rhair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(vpd2rhair(e1, T1) * 100., 2)),
            [res2, res2])

        # mixed types
        e1 = [vpd, vpd]
        T1 = tuple([Ta, Ta])
        assert isinstance(vpd2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, res2])
        e1 = tuple([vpd, vpd])
        T1 = [Ta, Ta]
        assert isinstance(vpd2rhair(e1, T1), tuple)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, res2])
        e1 = np.array(e1)
        T1 = [Ta, Ta]
        assert isinstance(vpd2rhair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(vpd2rhair(e1, T1) * 100., 2)),
            [res2, res2])
        e1 = [vpd, vpd]
        T1 = np.array(T1)
        assert isinstance(vpd2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        T1 = Ta
        assert isinstance(vpd2rhair(e1, T1, undef=-9998.), float)
        assert np.around(vpd2rhair(e1, T1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        T1 = Ta
        assert isinstance(vpd2rhair(e1, T1), float)
        assert np.around(vpd2rhair(e1, T1), 2) == -9999.
        e1 = vpd
        T1 = -9999.
        assert isinstance(vpd2rhair(e1, T1), float)
        assert np.around(vpd2rhair(e1, T1), 2) == -9999.

        # undef, list
        e1 = [vpd, -9999.]
        T1 = [Ta, Ta]
        assert isinstance(vpd2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, -9999. * 100.])
        e1 = [vpd, vpd]
        T1 = [Ta, -9999.]
        assert isinstance(vpd2rhair(e1, T1), list)
        self.assertEqual(
            list(np.around([ ee * 100. for ee in vpd2rhair(e1, T1)], 2)),
            [res2, -9999. * 100.])

        # undef, masked array
        e1 = np.ma.array([vpd, -9998., -9999.])
        T1 = [Ta, Ta, Ta]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(vpd2rhair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(vpd2rhair(e1, T1) * 100., 2)),
            [res2, masked, masked])
        e1 = [vpd, vpd, vpd]
        T1 = np.ma.array([Ta, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(vpd2rhair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(vpd2rhair(e1, T1) * 100., 2)),
            [res2, masked, masked])

    def test_eair2shair(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import eair2shair

        eair = 1000.
        Pa = 101325.
        res2 = 6.16

        # scalar
        e1 = eair
        P1 = Pa
        assert isinstance(eair2shair(e1, P1), float)
        assert np.around(eair2shair(e1, P1) * 1000., 2) == res2

        # list
        e1 = [eair, eair]
        P1 = Pa
        assert isinstance(eair2shair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, res2])
        e1 = eair
        P1 = [Pa, Pa]
        assert isinstance(eair2shair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, res2])
        e1 = [eair, eair]
        P1 = [Pa, Pa]
        assert isinstance(eair2shair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        P1 = tuple(P1)
        assert isinstance(eair2shair(e1, P1), tuple)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        P1 = np.array(P1)
        assert isinstance(eair2shair(e1, P1), np.ndarray)
        self.assertEqual(
            list(np.around(eair2shair(e1, P1) * 1000., 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        P1 = np.ma.array(P1)
        assert isinstance(eair2shair(e1, P1), np.ndarray)
        assert isinstance(eair2shair(e1, P1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2shair(e1, P1) * 1000., 2)),
            [res2, res2])

        # mixed types
        e1 = [eair, eair]
        P1 = tuple([Pa, Pa])
        assert isinstance(eair2shair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, res2])
        e1 = tuple([eair, eair])
        P1 = [Pa, Pa]
        assert isinstance(eair2shair(e1, P1), tuple)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, res2])
        e1 = np.array(e1)
        P1 = [Pa, Pa]
        assert isinstance(eair2shair(e1, P1), np.ndarray)
        self.assertEqual(
            list(np.around(eair2shair(e1, P1) * 1000., 2)),
            [res2, res2])
        e1 = [eair, eair]
        P1 = np.array(P1)
        assert isinstance(eair2shair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        P1 = Pa
        assert isinstance(eair2shair(e1, P1, undef=-9998.), float)
        assert np.around(eair2shair(e1, P1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        P1 = Pa
        assert isinstance(eair2shair(e1, P1), float)
        assert np.around(eair2shair(e1, P1), 2) == -9999.
        e1 = eair
        P1 = -9999.
        assert isinstance(eair2shair(e1, P1), float)
        assert np.around(eair2shair(e1, P1), 2) == -9999.

        # undef, list
        e1 = [eair, -9999.]
        P1 = [Pa, Pa]
        assert isinstance(eair2shair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, -9999. * 1000.])
        e1 = [eair, eair]
        P1 = [Pa, -9999.]
        assert isinstance(eair2shair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2shair(e1, P1)], 2)),
            [res2, -9999. * 1000.])

        # undef, masked array
        e1 = np.ma.array([eair, -9998., -9999.])
        P1 = [Pa, Pa, Pa]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(eair2shair(e1, P1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2shair(e1, P1) * 1000., 2)),
            [res2, masked, masked])
        e1 = [eair, eair, eair]
        P1 = np.ma.array([Pa, -9998., -9999.])
        P1 = np.ma.array(P1, mask=(P1 == -9998.))
        assert isinstance(eair2shair(e1, P1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2shair(e1, P1) * 1000., 2)),
            [res2, masked, masked])

    def test_shair2eair(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import shair2eair

        shair = 0.006
        Pa = 101325.
        res2 = 973.89

        # scalar
        e1 = shair
        T1 = Pa
        assert isinstance(shair2eair(e1, T1), float)
        assert np.around(shair2eair(e1, T1), 2) == res2

        # list
        e1 = [shair, shair]
        T1 = Pa
        assert isinstance(shair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = shair
        T1 = [Pa, Pa]
        assert isinstance(shair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = [shair, shair]
        T1 = [Pa, Pa]
        assert isinstance(shair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        T1 = tuple(T1)
        assert isinstance(shair2eair(e1, T1), tuple)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        T1 = np.array(T1)
        assert isinstance(shair2eair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        T1 = np.ma.array(T1)
        assert isinstance(shair2eair(e1, T1), np.ndarray)
        assert isinstance(shair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])

        # mixed types
        e1 = [shair, shair]
        T1 = tuple([Pa, Pa])
        assert isinstance(shair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = tuple([shair, shair])
        T1 = [Pa, Pa]
        assert isinstance(shair2eair(e1, T1), tuple)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = np.array(e1)
        T1 = [Pa, Pa]
        assert isinstance(shair2eair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = [shair, shair]
        T1 = np.array(T1)
        assert isinstance(shair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        T1 = Pa
        assert isinstance(shair2eair(e1, T1, undef=-9998.), float)
        assert np.around(shair2eair(e1, T1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        T1 = Pa
        assert isinstance(shair2eair(e1, T1), float)
        assert np.around(shair2eair(e1, T1), 2) == -9999.
        e1 = shair
        T1 = -9999.
        assert isinstance(shair2eair(e1, T1), float)
        assert np.around(shair2eair(e1, T1), 2) == -9999.

        # undef, list
        e1 = [shair, -9999.]
        T1 = [Pa, Pa]
        assert isinstance(shair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, -9999.])
        e1 = [shair, shair]
        T1 = [Pa, -9999.]
        assert isinstance(shair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, -9999.])

        # undef, masked array
        e1 = np.ma.array([shair, -9998., -9999.])
        T1 = [Pa, Pa, Pa]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(shair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, masked, masked])
        e1 = [shair, shair, shair]
        T1 = np.ma.array([Pa, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(shair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(shair2eair(e1, T1), 2)),
            [res2, masked, masked])

    def test_eair2mrair(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import eair2mrair

        eair = 1000.
        Pa = 101325.
        res2 = 6.2

        # scalar
        e1 = eair
        P1 = Pa
        assert isinstance(eair2mrair(e1, P1), float)
        assert np.around(eair2mrair(e1, P1) * 1000., 2) == res2

        # list
        e1 = [eair, eair]
        P1 = Pa
        assert isinstance(eair2mrair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, res2])
        e1 = eair
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, res2])
        e1 = [eair, eair]
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        P1 = tuple(P1)
        assert isinstance(eair2mrair(e1, P1), tuple)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        P1 = np.array(P1)
        assert isinstance(eair2mrair(e1, P1), np.ndarray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1) * 1000., 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        P1 = np.ma.array(P1)
        assert isinstance(eair2mrair(e1, P1), np.ndarray)
        assert isinstance(eair2mrair(e1, P1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1) * 1000., 2)),
            [res2, res2])

        # mixed types
        e1 = [eair, eair]
        P1 = tuple([Pa, Pa])
        assert isinstance(eair2mrair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, res2])
        e1 = tuple([eair, eair])
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1), tuple)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, res2])
        e1 = np.array(e1)
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1), np.ndarray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1) * 1000., 2)),
            [res2, res2])
        e1 = [eair, eair]
        P1 = np.array(P1)
        assert isinstance(eair2mrair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        P1 = Pa
        assert isinstance(eair2mrair(e1, P1, undef=-9998.), float)
        assert np.around(eair2mrair(e1, P1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        P1 = Pa
        assert isinstance(eair2mrair(e1, P1), float)
        assert np.around(eair2mrair(e1, P1), 2) == -9999.
        e1 = eair
        P1 = -9999.
        assert isinstance(eair2mrair(e1, P1), float)
        assert np.around(eair2mrair(e1, P1), 2) == -9999.

        # undef, list
        e1 = [eair, -9999.]
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, -9999. * 1000.])
        e1 = [eair, eair]
        P1 = [Pa, -9999.]
        assert isinstance(eair2mrair(e1, P1), list)
        self.assertEqual(
            list(np.around([ ee * 1000. for ee in eair2mrair(e1, P1)], 2)),
            [res2, -9999. * 1000.])

        # undef, masked array
        e1 = np.ma.array([eair, -9998., -9999.])
        P1 = [Pa, Pa, Pa]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(eair2mrair(e1, P1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1) * 1000., 2)),
            [res2, masked, masked])
        e1 = [eair, eair, eair]
        P1 = np.ma.array([Pa, -9998., -9999.])
        P1 = np.ma.array(P1, mask=(P1 == -9998.))
        assert isinstance(eair2mrair(e1, P1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1) * 1000., 2)),
            [res2, masked, masked])

        # mol = True

        eair = 1000.
        Pa = 101325.
        res2 = 9.97

        # scalar
        e1 = eair
        P1 = Pa
        assert isinstance(eair2mrair(e1, P1, mol=True), float)
        assert np.around(eair2mrair(e1, P1, mol=True) * 1000., 2) == res2

        # list
        e1 = [eair, eair]
        P1 = Pa
        assert isinstance(eair2mrair(e1, P1, mol=True), list)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, res2])
        e1 = eair
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1, mol=True), list)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, res2])
        e1 = [eair, eair]
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1, mol=True), list)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        P1 = tuple(P1)
        assert isinstance(eair2mrair(e1, P1, mol=True), tuple)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        P1 = np.array(P1)
        assert isinstance(eair2mrair(e1, P1, mol=True), np.ndarray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1, mol=True) * 1000., 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        P1 = np.ma.array(P1)
        assert isinstance(eair2mrair(e1, P1, mol=True), np.ndarray)
        assert isinstance(eair2mrair(e1, P1, mol=True), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1, mol=True) * 1000., 2)),
            [res2, res2])

        # mixed types
        e1 = [eair, eair]
        P1 = tuple([Pa, Pa])
        assert isinstance(eair2mrair(e1, P1, mol=True), list)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, res2])
        e1 = tuple([eair, eair])
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1, mol=True), tuple)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, res2])
        e1 = np.array(e1)
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1, mol=True), np.ndarray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1, mol=True) * 1000., 2)),
            [res2, res2])
        e1 = [eair, eair]
        P1 = np.array(P1)
        assert isinstance(eair2mrair(e1, P1, mol=True), list)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        P1 = Pa
        assert isinstance(eair2mrair(e1, P1, undef=-9998.), float)
        assert np.around(eair2mrair(e1, P1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        P1 = Pa
        assert isinstance(eair2mrair(e1, P1, mol=True), float)
        assert np.around(eair2mrair(e1, P1, mol=True), 2) == -9999.
        e1 = eair
        P1 = -9999.
        assert isinstance(eair2mrair(e1, P1, mol=True), float)
        assert np.around(eair2mrair(e1, P1, mol=True), 2) == -9999.

        # undef, list
        e1 = [eair, -9999.]
        P1 = [Pa, Pa]
        assert isinstance(eair2mrair(e1, P1, mol=True), list)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, -9999. * 1000.])
        e1 = [eair, eair]
        P1 = [Pa, -9999.]
        assert isinstance(eair2mrair(e1, P1, mol=True), list)
        self.assertEqual(
            list(np.around([ ee * 1000.
                             for ee in eair2mrair(e1, P1, mol=True)], 2)),
            [res2, -9999. * 1000.])

        # undef, masked array
        e1 = np.ma.array([eair, -9998., -9999.])
        P1 = [Pa, Pa, Pa]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(eair2mrair(e1, P1, mol=True), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1, mol=True) * 1000., 2)),
            [res2, masked, masked])
        e1 = [eair, eair, eair]
        P1 = np.ma.array([Pa, -9998., -9999.])
        P1 = np.ma.array(P1, mask=(P1 == -9998.))
        assert isinstance(eair2mrair(e1, P1, mol=True), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(eair2mrair(e1, P1, mol=True) * 1000., 2)),
            [res2, masked, masked])

    def test_mrair2eair(self):
        import numpy as np
        from numpy.ma import masked
        from pyjams import mrair2eair

        mrair = 0.006
        Pa = 101325.
        res2 = 968.10

        # scalar
        e1 = mrair
        T1 = Pa
        assert isinstance(mrair2eair(e1, T1), float)
        assert np.around(mrair2eair(e1, T1), 2) == res2

        # list
        e1 = [mrair, mrair]
        T1 = Pa
        assert isinstance(mrair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = mrair
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = [mrair, mrair]
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        T1 = tuple(T1)
        assert isinstance(mrair2eair(e1, T1), tuple)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        T1 = np.array(T1)
        assert isinstance(mrair2eair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        T1 = np.ma.array(T1)
        assert isinstance(mrair2eair(e1, T1), np.ndarray)
        assert isinstance(mrair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])

        # mixed types
        e1 = [mrair, mrair]
        T1 = tuple([Pa, Pa])
        assert isinstance(mrair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = tuple([mrair, mrair])
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1), tuple)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = np.array(e1)
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1), np.ndarray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])
        e1 = [mrair, mrair]
        T1 = np.array(T1)
        assert isinstance(mrair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        T1 = Pa
        assert isinstance(mrair2eair(e1, T1, undef=-9998.), float)
        assert np.around(mrair2eair(e1, T1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        T1 = Pa
        assert isinstance(mrair2eair(e1, T1), float)
        assert np.around(mrair2eair(e1, T1), 2) == -9999.
        e1 = mrair
        T1 = -9999.
        assert isinstance(mrair2eair(e1, T1), float)
        assert np.around(mrair2eair(e1, T1), 2) == -9999.

        # undef, list
        e1 = [mrair, -9999.]
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, -9999.])
        e1 = [mrair, mrair]
        T1 = [Pa, -9999.]
        assert isinstance(mrair2eair(e1, T1), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, -9999.])

        # undef, masked array
        e1 = np.ma.array([mrair, -9998., -9999.])
        T1 = [Pa, Pa, Pa]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(mrair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, masked, masked])
        e1 = [mrair, mrair, mrair]
        T1 = np.ma.array([Pa, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(mrair2eair(e1, T1), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1), 2)),
            [res2, masked, masked])

        # mol = True

        mrair = 0.006
        Pa = 101325.
        res2 = 604.32

        # scalar
        e1 = mrair
        T1 = Pa
        assert isinstance(mrair2eair(e1, T1, mol=True), float)
        assert np.around(mrair2eair(e1, T1, mol=True), 2) == res2

        # list
        e1 = [mrair, mrair]
        T1 = Pa
        assert isinstance(mrair2eair(e1, T1, mol=True), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])
        e1 = mrair
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1, mol=True), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])
        e1 = [mrair, mrair]
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1, mol=True), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])

        # tuple
        e1 = tuple(e1)
        T1 = tuple(T1)
        assert isinstance(mrair2eair(e1, T1, mol=True), tuple)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])

        # ndarray
        e1 = np.array(e1)
        T1 = np.array(T1)
        assert isinstance(mrair2eair(e1, T1, mol=True), np.ndarray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])

        # masked_array
        e1 = np.ma.array(e1)
        T1 = np.ma.array(T1)
        assert isinstance(mrair2eair(e1, T1, mol=True), np.ndarray)
        assert isinstance(mrair2eair(e1, T1, mol=True), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])

        # mixed types
        e1 = [mrair, mrair]
        T1 = tuple([Pa, Pa])
        assert isinstance(mrair2eair(e1, T1, mol=True), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])
        e1 = tuple([mrair, mrair])
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1, mol=True), tuple)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])
        e1 = np.array(e1)
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1, mol=True), np.ndarray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])
        e1 = [mrair, mrair]
        T1 = np.array(T1)
        assert isinstance(mrair2eair(e1, T1, mol=True), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, res2])

        # undef, scalar
        e1 = -9998.
        T1 = Pa
        assert isinstance(mrair2eair(e1, T1, undef=-9998.), float)
        assert np.around(mrair2eair(e1, T1, undef=-9998.), 2) == -9998.
        e1 = -9999.
        T1 = Pa
        assert isinstance(mrair2eair(e1, T1, mol=True), float)
        assert np.around(mrair2eair(e1, T1, mol=True), 2) == -9999.
        e1 = mrair
        T1 = -9999.
        assert isinstance(mrair2eair(e1, T1, mol=True), float)
        assert np.around(mrair2eair(e1, T1, mol=True), 2) == -9999.

        # undef, list
        e1 = [mrair, -9999.]
        T1 = [Pa, Pa]
        assert isinstance(mrair2eair(e1, T1, mol=True), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, -9999.])
        e1 = [mrair, mrair]
        T1 = [Pa, -9999.]
        assert isinstance(mrair2eair(e1, T1, mol=True), list)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, -9999.])

        # undef, masked array
        e1 = np.ma.array([mrair, -9998., -9999.])
        T1 = [Pa, Pa, Pa]
        e1 = np.ma.array(e1, mask=(e1 == -9998.))
        assert isinstance(mrair2eair(e1, T1, mol=True), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, masked, masked])
        e1 = [mrair, mrair, mrair]
        T1 = np.ma.array([Pa, -9998., -9999.])
        T1 = np.ma.array(T1, mask=(T1 == -9998.))
        assert isinstance(mrair2eair(e1, T1, mol=True), np.ma.MaskedArray)
        self.assertEqual(
            list(np.around(mrair2eair(e1, T1, mol=True), 2)),
            [res2, masked, masked])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Convert integer to and from Roman numerals

This module was written by Matthias Cuntz while at Department of Computational
Hydrosystems, Helmholtz Centre for Environmental Research - UFZ, Leipzig,
Germany, and continued while at Institut National de Recherche pour
l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2012-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   int2roman
   roman2int

Notes
-----
Code adapted from Tim Valenta:
http://code.activestate.com/recipes/81611-roman-numerals


History
-------
    * Written May 2012 by Matthias Cuntz (mc (at) macu (dot) de)
    * Option lower in int2roman, May 2012, Matthias Cuntz
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Use assert mechanism, Apr 2014, Matthias Cuntz
    * Make numpy docstring format, Nov 2021, Matthias Cuntz
    * Ported into pyjams, Nov 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""


__all__ = ['int2roman', 'roman2int']


_numeral_map = list(zip((1000, 900, 500, 400, 100, 90,
                         50, 40, 10, 9, 5, 4, 1),
                        ('M', 'CM', 'D', 'CD', 'C', 'XC',
                         'L', 'XL', 'X', 'IX', 'V', 'IV', 'I')))


def int2roman(i, lower=False):
    """
    Convert an integer to a Roman numeral

    Parameters
    ----------
    i : int
        Integer number to convert
    lower : bool, optional
        Output lowercase numerals if True,
        else output uppercase numerals (default)

    Returns
    -------
    str
        Roman numeral

    Notes
    -----
    Code adapted from Tim Valenta:
    http://code.activestate.com/recipes/81611-roman-numerals

    Examples
    --------
    >>> print(int2roman(1))
    I
    >>> print(int2roman(19))
    XIX
    >>> print(int2roman(159))
    CLIX
    >>> print(int2roman(159, lower=True))
    clix

    """
    assert i >= 1, 'integer must be > 0.'

    result = []
    for integer, numeral in _numeral_map:
        count = int(i // integer)
        result.append(numeral * count)
        i -= integer * count

    if lower:
        result = [ i.lower() for i in result ]

    return ''.join(result)


def roman2int(n):
    """
    Convert a Roman numeral to an integer

    Parameters
    ----------
    i : str
        String with Roman numeral to convert to integer

    Returns
    -------
    Int
        Integer

    Notes
    -----
    Input can only be one single numeral.

    Code adapted from Tim Valenta:
    http://code.activestate.com/recipes/81611-roman-numerals

    Examples
    --------
    >>> print(roman2int('I'))
    1
    >>> print(roman2int('i'))
    1
    >>> print(roman2int('iv'))
    4
    >>> print(roman2int('MCCCLIV'))
    1354

    """
    n = str(n).upper()
    result = 0
    i = 0
    for integer, numeral in _numeral_map:
        while n[i:i + len(numeral)] == numeral:
            result += integer
            i += len(numeral)

    return result


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
Area of grid cells on Earth

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2009-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   gridcellarea

History
    * Written Jul 2009 by Matthias Cuntz - mc (at) macu (dot) de
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Assert correct input, Apr 2014, Matthias Cuntz
    * Clip lat because -90 and 90 give negative area, Oct 2018, Matthias Cuntz
    * Corrected bug in checking ascending or descending lats,
      Apr 2022, Matthias Cuntz
    * Removed assertion of nlat/nlon > 2, Apr 2022, Matthias Cuntz
    * Keyword radius of Earth and changed default from 6371000 to 6371009
      as in the rest of pyjams, Apr 2022, Matthias Cuntz
    * Rename cellarea to gridcellarea, Apr 2022, Matthias Cuntz

"""
import numpy as np


__all__ = ['gridcellarea']


def gridcellarea(lat, lon, globe=False, rearth=6371009.):
    """
    Area of grid cells on a spherical Earth in square metre

    Parameters
    ----------
    lat : array_like
        Latitudes of grid cell centres in degrees N
    lon : array_like
        Longitudes of grid cell centres in degrees E
    globe : bool, optional
        Assumes that latitudes span the globe if True,
        i.e. they are bounded by 90 and -90 degrees latitude
    rearth : float, optional
        Radius of the spherical Earth (default: 6371009.)

    Returns
    -------
    numpy array with area in m^2


    Notes
    -----
    This is a rather rough routine with lots of possible improvements,
    notably allowing irregular grids, grid boundaries, etc.


    Examples
    --------

    Gaussian latitudes

    >>> lat = np.array([ 12.98898858, 9.27785325, 5.56671363])
    >>> lon = np.array([ 0., 3.75, 7.5])

    >>> print(gridcellarea(lat, lon)[0,:])
    [1.67639557e+11 1.67639557e+11 1.67639557e+11]

    >>> print(gridcellarea(lat, lon)[1,:])
    [1.69790907e+11 1.69790907e+11 1.69790907e+11]

    >>> print(gridcellarea(lat, lon)[2,:])
    [1.71230373e+11 1.71230373e+11 1.71230373e+11]

    """
    # assert numpy
    lati = np.array(lat)
    loni = np.array(lon)
    nlat = lati.size
    nlon = loni.size

    # assert -90 < lat < 90
    assert np.abs(lat).max() <= 90., (
        'probably swapped lat and lon in call:'
        ' def gridcellarea(lat, lon, globe=False):')

    # lat size in degrees
    # still + or -
    dlat = lati - np.roll(lati, 1)
    if globe:
        if dlat[0] > 0:  # descending lats
            l0 = 90.
        else:            # ascending lats
            l0 = -90.
        dlat[0]  = (lati[0] - l0) + 0.5 * (lati[1] - lati[0])
        dlat[-1] = (-l0 - lati[-1]) - 0.5*(lati[-2] - lati[-1])
    else:
        dlat[0]  = lati[1] - lati[0]
        dlat[-1] = lati[-1] - lati[-2]

    # lon size in degrees
    # check if meridian in lon range -> shift to -180,180
    # e.g. 358 359 0 1
    if np.any(np.abs(np.diff(loni)) > 360. / nlon):
        loni = np.where(loni > 180., loni - 360., loni)
    # check if -180,180 longitude in lon range -> shift to 0,360
    # e.g. 179 180 -179 -178
    if np.any(np.abs(np.diff(loni)) > 360. / nlon):
        loni = np.where(loni < 0., loni + 360., loni)
    dlon = np.abs(loni - np.roll(loni, 1))
    dlon[0] = np.abs(loni[1] - loni[0])

    # Northern latitudes of grid cell edges
    n_lat = lati[0] + dlat[0]/2. + np.cumsum(dlat)

    # Area of grid cells in m^2 with lat/lon in degree
    d2r = np.pi / 180.  # degree to radian

    area = np.empty([nlat, nlon])
    dlat = np.abs(dlat[:])
    # -90, 90 give negative np.cos(lat*d2r)
    lati = np.clip(lati, -89.99999, 89.99999)
    for i in range(nlon):
        area[:, i] = (2. * d2r * rearth**2 * dlon[i] *
                      np.sin(0.5 * dlat * d2r) * np.cos(lati * d2r))

    return area


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
